/* ====================================================================
 *  Copyright 2004 Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_protocol.h"
#include "http_log.h"

#include "apr_strings.h"

#define MOD_LOAD_AVERAGE_VERSION "0.1.0"

module AP_MODULE_DECLARE_DATA load_average_module;

typedef struct
{
    float max_load_average;
    apr_hash_t *la;
} load_average_config_rec;

typedef struct
{
    float value;
} la_hash_value;


static int mod_load_average_handler(request_rec * r)
{
    la_hash_value *t;
    load_average_config_rec *conf =
        (load_average_config_rec *) ap_get_module_config(r->server->
                                                         module_config,
                                                         &load_average_module);
    /**
     * We Do not Handle Sub Requests.  
     * It is assumed that if we want to return a 503, 
     * we will do it in the main request, not a subrequest. 
     */
    if (HTTP_SERVICE_UNAVAILABLE == r->status) {
        return DECLINED;
    }

    if (!r->handler) {
        ap_log_rerror(APLOG_MARK, APLOG_WARNING, 0, r,
                      "LoadAverage: No Handler was set?");
        return DECLINED;
    }

    t = apr_hash_get(conf->la, r->handler, strlen(r->handler));

    if (t != NULL) {
        double loadavg[1];
        int ret;
        ret = getloadavg(loadavg, 1);

        if (ret < 0) {
            ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
                          "getloadavg() returned: '%d'. We wanted a return code greater than 0.",
                          ret);
            return DECLINED;
        }

        if (loadavg[0] > t->value) {
            return HTTP_SERVICE_UNAVAILABLE;
        }

    }

    return DECLINED;
}

static int mod_load_average_post_read(request_rec * r)
{
    load_average_config_rec *conf =
        (load_average_config_rec *) ap_get_module_config(r->server->
                                                         module_config,
                                                         &load_average_module);
    if (HTTP_SERVICE_UNAVAILABLE == r->status) {
        return DECLINED;
    }

    if (conf && conf->max_load_average > 0) {
        double loadavg[1];
        int ret;
        ret = getloadavg(loadavg, 1);
        if (ret < 0) {
            ap_log_rerror(APLOG_MARK, APLOG_NOTICE, 0, r,
                          "getloadavg() returned: '%d'. We wanted a return code greater than 0.",
                          ret);
            return DECLINED;
        }

        if (loadavg[0] > conf->max_load_average) {
            return HTTP_SERVICE_UNAVAILABLE;
        }
    }

    return DECLINED;
}

static void mod_load_average_register_hooks(apr_pool_t * p)
{
    ap_hook_post_read_request(mod_load_average_post_read, NULL, NULL,
                              APR_HOOK_FIRST);
    ap_hook_handler(mod_load_average_handler, NULL, NULL, APR_HOOK_FIRST);
}


static const char *load_average_max_set(cmd_parms * parms, void *config,
                                        const char *value)
{
    char *bp;
    load_average_config_rec *conf =
        (load_average_config_rec *) ap_get_module_config(parms->server->
                                                         module_config,
                                                         &load_average_module);

    conf->max_load_average = strtof(value, &bp);
    if (conf->max_load_average < 0.0) {
        return "LoadAvgeMax Must be greater than 0!";
    }
    return NULL;
}

static const char *load_average_handler_set(cmd_parms * parms, void *config,
                                            const char *key,
                                            const char *value)
{
    float f;
    char *bp;
    la_hash_value *t;
    load_average_config_rec *conf =
        (load_average_config_rec *) ap_get_module_config(parms->server->
                                                         module_config,
                                                         &load_average_module);


    f = strtof(value, &bp);
    if (f < 0.0) {
        return " Must LoadAvgMaxByHandler be greater than 0!";
    }
    t = apr_palloc(parms->pool, sizeof(la_hash_value));
    t->value = f;
    apr_hash_set(conf->la, key, strlen(key), t);
    return NULL;
}


static const command_rec load_average_cmds[] = {
    AP_INIT_TAKE1("LoadAvgMax", load_average_max_set,
                  NULL,
                  RSRC_CONF,
                  "Maximum Load Average."),
    AP_INIT_TAKE2("LoadAvgMaxByHandler", load_average_handler_set,
                  NULL,
                  RSRC_CONF,
                  "Maximum Load Average Per Handler"),
    {NULL}
};


static void *create_load_average_config(apr_pool_t * p, server_rec * s)
{
    load_average_config_rec *conf;

    conf =
        (load_average_config_rec *) apr_pcalloc(p,
                                                sizeof
                                                (load_average_config_rec));
    if (conf) {
        conf->max_load_average = 100.0;
        conf->la = apr_hash_make(p);
    }

    return conf;
}

module AP_MODULE_DECLARE_DATA load_average_module = {
    STANDARD20_MODULE_STUFF,
    NULL,
    NULL,
    create_load_average_config,
    NULL,
    load_average_cmds,
    mod_load_average_register_hooks,
};
